from json import dumps
from typing import List, Optional, Tuple
from urllib.parse import parse_qsl, urlencode

from spylib.hmac import validate


def validate_signed_query_string(query_string: str, *, api_secret_key: str):
    """Validates that a query string has been signed by Shopify.

    [Implements the parsing algorithm defined by Shopify here](https://shopify.dev/apps/auth/oauth/getting-started#step-7-verify-a-request).
    Including the special case [`ids` parameter parsing](https://shopify.dev/apps/auth/oauth/getting-started#ids-array-parameter)


    Args:
        query_string: A valid query string. `hmac=123&test=456`
        api_secret_key: The api secret key from Shopify partners.

    Raises:
        Exception: `ValueError`
    """

    signature: Optional[str] = None
    query_params: List[Tuple[str, str]] = []

    # I tried a number of options here to avoid doing
    # manual parsing and string manipulation. `parse_qs`
    # can work as long as you pass `doseq=True` & `quote_via=quote``
    # to urlencode. It does not, however, handle the ids param.
    # The code ended up being cleaner this way in one loop.

    ids: List[str] = []
    for key, value in parse_qsl(query_string, strict_parsing=True):
        if key == 'hmac':
            signature = value
            continue
        if key == 'ids[]':
            if not ids:
                query_params.append(('ids', ''))
            ids.append(value)
            continue
        query_params.append((key, value))

    # `safe` param via: https://stackoverflow.com/a/49244224
    message = urlencode(query_params, safe=':/').replace('ids=', f'ids={dumps(ids)}')

    validate(
        sent_hmac=signature or '',
        message=message,
        secret=api_secret_key,
    )



from datetime import datetime, timezone

from .shortuuid import random


def now_epoch() -> int:
    return int(datetime.now(timezone.utc).timestamp())


def get_unique_id() -> str:
    return random(length=10)



from typing import List

from pydantic import BaseModel, validator


def _parse_scope(scope: str) -> List[str]:
    return scope.split(',')


class OfflineTokenModel(BaseModel):
    """[Read more about Offline access](https://shopify.dev/apps/auth/oauth/access-modes#offline-access)."""

    access_token: str
    """
    An API access token that can be used to access the shop's data as long as your app is installed. Your app should store the token somewhere to make authenticated requests for a shop’s data.
    """

    scope: List[str]
    """
    The list of access scopes that were granted to your app and are associated with the access token.
    """

    _normalize_scope = validator('scope', allow_reuse=True, pre=True)(_parse_scope)


class AssociatedUser(BaseModel):
    """Shopify staff user associated with an online token."""

    id: int
    """
    The user id.

    [Can be associated with the `sub` of a session token](https://shopify.dev/apps/auth/oauth/session-tokens#payload)
    """

    first_name: str
    """
    The user's first name.
    """

    last_name: str
    """
    The user's last name.
    """

    email: str
    """
    The user's email.
    """

    email_verified: bool = False
    """
    If the user's email has been verified.

    If `False` email cannot be trusted (legacy accounts).
    """

    account_owner: bool
    """
    Whether the user is the owner of the Shopify account.
    """

    locale: str
    """
    The user's preferred locale. Locale values use the format `language` or `language-COUNTRY`, where `language` is a two-letter language code, and `COUNTRY` is a two-letter country code. For example: `en` or `en-US`.
    """

    collaborator: bool
    """
    If the user is a partner who collaborates with the merchant.
    """


class OnlineTokenModel(BaseModel):
    """[Read more about Online access](https://shopify.dev/apps/auth/oauth/access-modes#online-access)."""

    access_token: str
    """An API access token that can be used to access the shop's data until it expires or the associated user logs out."""

    scope: List[str]
    """The list of access scopes that were requested. Inspect `associated_user_scope` to see which were granted."""

    expires_in: int
    """The number of seconds until this session (and `access_token`) expire."""

    associated_user_scope: List[str]
    """The list of access scopes that were both requested and available to this user."""

    associated_user: AssociatedUser
    """The Shopify user associated with this token."""

    _normalize_scope = validator('scope', allow_reuse=True, pre=True)(_parse_scope)
    _normalize_associated_user_scope = validator(
        'associated_user_scope', allow_reuse=True, pre=True
    )(_parse_scope)



from dataclasses import dataclass
from typing import Awaitable, Callable, List, Optional

from spylib.exceptions import FastAPIImportError

try:
    from fastapi import APIRouter, Depends, HTTPException, Query  # type: ignore
except ImportError as e:
    raise FastAPIImportError(
        'The oauth router is a fastapi router and fastapi is not installed. '
        'Run `pip install spylib[fastapi]` to be able to use the oauth router.'
    ) from e

from starlette.requests import Request
from starlette.responses import RedirectResponse

from ..utils import store_domain
from .callback import process_callback
from .models import OfflineTokenModel, OnlineTokenModel
from .redirects import app_redirect, oauth_init_url


@dataclass
class Callback:
    code: str = Query(...)
    hmac: str = Query(...)
    timestamp: int = Query(...)
    state: str = Query(...)
    shop: str = Query(...)


def init_oauth_router(
    app_scopes: List[str],
    user_scopes: List[str],
    public_domain: str,
    private_key: str,
    api_key: str,
    api_secret_key: str,
    post_install: Callable[[str, OfflineTokenModel], Optional[Awaitable]],
    post_login: Optional[Callable[[str, OnlineTokenModel], Optional[Awaitable]]] = None,
    install_init_path='/shopify/auth',
    callback_path='/callback',
    path_prefix: str = '',
) -> APIRouter:
    router = APIRouter()

    if not install_init_path.startswith('/'):
        raise ValueError('The install_init_path argument must start with "/"')
    if not callback_path.startswith('/'):
        raise ValueError('The callback_path argument must start with "/"')
    if path_prefix and not path_prefix.startswith('/'):
        raise ValueError('The path_prefix argument must start with "/"')

    @router.get(install_init_path, include_in_schema=False)
    async def shopify_auth(shop: str):
        """Endpoint initiating the OAuth process on a Shopify store."""
        return RedirectResponse(
            oauth_init_url(
                domain=store_domain(shop=shop),
                is_login=False,
                requested_scopes=app_scopes,
                callback_domain=public_domain,
                callback_path=callback_path,
                path_prefix=path_prefix,
                jwt_key=private_key,
                api_key=api_key,
            )
        )

    @router.get(callback_path, include_in_schema=False)
    async def shopify_callback(request: Request, shop: str, args: Callback = Depends(Callback)):
        """REST endpoint called by Shopify during the OAuth process for installation and login."""
        try:
            oauthjwt = await process_callback(
                shop=args.shop,
                timestamp=args.timestamp,
                query_string=request.scope['query_string'],
                api_secret_key=api_secret_key,
                api_key=api_key,
                state=args.state,
                private_key=private_key,
                code=args.code,
                post_install=post_install,
                post_login=post_login,
            )
        except Exception as e:
            raise HTTPException(status_code=400, detail=f'Validation failed: {e}')

        # === If installation ===
        # Setup the login obj and redirect to oauth_redirect
        if not oauthjwt.is_login:
            if post_login is None:
                return RedirectResponse(
                    app_redirect(
                        store_domain=args.shop,
                        app_domain=public_domain,
                        app_api_key=api_key,
                    )
                )
            # Initiate the oauth loop for login
            return RedirectResponse(
                oauth_init_url(
                    domain=args.shop,
                    is_login=True,
                    requested_scopes=user_scopes,
                    callback_domain=public_domain,
                    callback_path=callback_path,
                    path_prefix=path_prefix,
                    jwt_key=private_key,
                    api_key=api_key,
                )
            )

        # Redirect to the app in Shopify admin
        return RedirectResponse(
            app_redirect(
                store_domain=args.shop,
                app_domain=public_domain,
                app_api_key=api_key,
            )
        )

    return router



from typing import Optional

from fastapi import Request, Security  # type: ignore
from fastapi.openapi.models import APIKey, APIKeyIn  # type: ignore
from fastapi.security.api_key import APIKeyBase  # type: ignore
from starlette.exceptions import HTTPException
from starlette.status import HTTP_401_UNAUTHORIZED, HTTP_403_FORBIDDEN

from spylib import webhook

SHOPIFY_WEBHOOK_HMAC_HEADER = 'X-Shopify-Hmac-SHA256'


class WebhookHMACHeader(APIKeyBase):
    """Reuse APIKey to authenticate the webhook using the HMAC signature.

    It requires our own implementation of APIKeyHeader because we need to
    pass in the api secret and process the whole request object.
    """

    def __init__(
        self,
        *,
        name: str,
        api_secret_key: str,
        scheme_name: Optional[str] = None,
        description: Optional[str] = None,
        auto_error: bool = True,
    ):
        self.model: APIKey = APIKey(**{'in': APIKeyIn.query}, name=name, description=description)
        self.scheme_name = scheme_name or self.__class__.__name__
        self.auto_error = auto_error
        self.api_secret_key = api_secret_key

    async def __call__(self, request: Request) -> bool:
        data = str(await request.body(), 'utf-8')
        if not self.api_secret_key:
            raise HTTPException(
                status_code=HTTP_403_FORBIDDEN, detail='api_secret_key must be set'
            )

        hmac_header = request.headers.get(SHOPIFY_WEBHOOK_HMAC_HEADER, '')
        return webhook.validate(
            data=data, hmac_header=hmac_header, api_secret_key=self.api_secret_key
        )


webhook_hmac = WebhookHMACHeader(name=SHOPIFY_WEBHOOK_HMAC_HEADER, api_secret_key='')


def authenticate_webhook_hmac(hmac: bool = Security(webhook_hmac)):
    """Dependency for authenticating the webhook using the HMAC signature.

    `webhook_hmac.api_secret_key` needs to be assigned with a valid secret prior to using this dependency.

    ```python
    webhook_hmac.api_secret_key = 'SOME_SECRET'
    @app.post('/SOME_PATH')
    def some_path(is_hmac_valid: bool = Depends(authenticate_webhook_hmac)):
        if not is_hmac_valid:
            return HTTPException(status_code=400, detail='Invalid HMAC')
    ```
    """

    if not hmac:
        raise HTTPException(
            status_code=HTTP_401_UNAUTHORIZED, detail='Webhook HMAC authentication failed'
        )
    else:
        return hmac



from typing import Union

from httpx import AsyncClient, codes

from .models import OfflineTokenModel, OnlineTokenModel


async def exchange_token(
    *,
    shop: str,
    code: str,
    api_key: str,
    api_secret_key: str,
) -> Union[OnlineTokenModel, OfflineTokenModel]:
    """[Exchanges the temporary authorization code with Shopify for a token](https://shopify.dev/apps/auth/oauth/getting-started#step-5-get-a-permanent-access-token).

    [All prior security checks must already have been completed.](https://shopify.dev/apps/auth/oauth/getting-started#step-4-confirm-installation)

    Args:
        shop (str): The name of the merchant's shop. `example.myshopify.com`
        code (str): The authorization code provided in the redirect.
        api_key (str): The API key for the app, as defined in the Shopify Partner Dashboard.
        api_secret_key (str): The API secret key for the app, as defined in the Shopify Partner Dashboard.

    Returns:
        Union[OnlineTokenModel, OfflineTokenModel]: Validated token response. Will differ [depending upon the requested access mode.](https://shopify.dev/apps/auth/oauth/access-modes)
    """

    async with AsyncClient(base_url=f'https://{shop}') as client:
        response = await client.post(
            url='/admin/oauth/access_token',
            json={
                'code': code,
                'client_id': api_key,
                'client_secret': api_secret_key,
            },
        )

    if response.status_code != codes.OK:
        message = (
            f'Shopify rejected token exchange. Shop: "{shop}". Status: "{response.status_code}".'
        )
        raise Exception(message)  # Not sure what the convention is here

    raw_response_body = response.json()
    is_online_token = 'associated_user' in raw_response_body

    if is_online_token:
        return OnlineTokenModel.parse_obj(raw_response_body)

    return OfflineTokenModel.parse_obj(raw_response_body)


async def exchange_offline_token(
    *,
    shop: str,
    code: str,
    api_key: str,
    api_secret_key: str,
) -> OfflineTokenModel:
    """[Exchanges the temporary authorization code with Shopify for an offline token](https://shopify.dev/apps/auth/oauth/getting-started#step-5-get-a-permanent-access-token).

    Can be used instead of `exchange_token` when you know by some mechanism it will only be an offline token.

    [All prior security checks must already have been completed.](https://shopify.dev/apps/auth/oauth/getting-started#step-4-confirm-installation)

    Args:
        shop (str): The name of the merchant's shop. `example.myshopify.com`
        code (str): The authorization code provided in the redirect.
        api_key (str): The API key for the app, as defined in the Shopify Partner Dashboard.
        api_secret_key (str): The API secret key for the app, as defined in the Shopify Partner Dashboard.

    Returns:
        OfflineTokenModel: [Validated offline token response.]https://shopify.dev/apps/auth/oauth/access-modes#offline-access
    """
    token = await exchange_token(
        shop=shop, code=code, api_key=api_key, api_secret_key=api_secret_key
    )

    assert isinstance(token, OfflineTokenModel)

    return token


async def exchange_online_token(
    *,
    shop: str,
    code: str,
    api_key: str,
    api_secret_key: str,
) -> OnlineTokenModel:
    """[Exchanges the temporary authorization code with Shopify for an online token](https://shopify.dev/apps/auth/oauth/getting-started#step-5-get-a-permanent-access-token).

    Can be used instead of `exchange_token` when you know by some mechanism it will only be an online token.

    [All prior security checks must already have been completed.](https://shopify.dev/apps/auth/oauth/getting-started#step-4-confirm-installation)

    Args:
        shop (str): The name of the merchant's shop. `example.myshopify.com`
        code (str): The authorization code provided in the redirect.
        api_key (str): The API key for the app, as defined in the Shopify Partner Dashboard.
        api_secret_key (str): The API secret key for the app, as defined in the Shopify Partner Dashboard.

    Returns:
        OnlineTokenModel: [Validated online token response.]https://shopify.dev/apps/auth/oauth/access-modes#online-access
    """
    token = await exchange_token(
        shop=shop, code=code, api_key=api_key, api_secret_key=api_secret_key
    )

    assert isinstance(token, OnlineTokenModel)

    return token



from enum import Enum

from pydantic import BaseModel
from starlette import status


class Method(Enum):
    GET = 'get'
    POST = 'post'
    PUT = 'put'
    DELETE = 'delete'


class Request(BaseModel):
    method: Method
    good_status: int


GET = Request(method=Method.GET, good_status=status.HTTP_200_OK)
POST = Request(method=Method.POST, good_status=status.HTTP_201_CREATED)
PUT = Request(method=Method.PUT, good_status=status.HTTP_200_OK)
DELETE = Request(method=Method.DELETE, good_status=status.HTTP_200_OK)



# Code copied over from shortuuid package:
# https://github.com/skorokithakis/shortuuid/blob/v1.0.8/shortuuid/main.py
import binascii
import os

ALPHABET = '23456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz'


def random(length):
    """Generate a cryptographically-secure short random string of the specified length."""
    random_num = int(binascii.b2a_hex(os.urandom(length)), 16)
    return int_to_string(random_num, ALPHABET, padding=length)[:length]


def int_to_string(number, alphabet, padding=None):
    """Convert a number to a string, using the given alphabet.

    The output has the most significant digit first.
    """
    output = ''
    alpha_len = len(alphabet)
    while number:
        number, digit = divmod(number, alpha_len)
        output += alphabet[digit]
    if padding:
        remainder = max(padding - len(output), 0)
        output = output + alphabet[0] * remainder
    return output[::-1]



import jwt
from pydantic import BaseModel, validator

from .misc import now_epoch


class JWTBaseModel(BaseModel):
    """Base class to manage JWT.

    The pydantic model fields are the data content of the JWT.
    The default expiration (exp) is set to 900 seconds. Overwrite the ClassVar exp to change the
    expiration time.
    """

    exp: int = None  # type: ignore

    @validator('exp', pre=True, always=True)
    def set_id(cls, exp):
        return exp or (now_epoch() + 900)

    @classmethod
    def decode_token(cls, key: str, token: str, verify: bool = True):
        """Decode the token and load the data content into an instance of this class.

        Args
        ----------
        key: Secret key used to encrypt the JWT
        verify: If true, verify the signature is valid, otherwise skip. Default is True

        Returns
        -------
        Class instance
        """
        data = jwt.decode(token, key, algorithms=['HS256'], verify=verify)
        # Enforce conversion to satisfy typing
        data = dict(data)
        return cls(**data)

    def encode_token(self, key: str) -> str:
        """Encode the class data into a JWT and return a string.

        Args
        ----------
        key: Secret key used to encrypt the JWT

        Returns
        -------
        The JWT as a string
        """
        data = self.dict()
        data['exp'] = self.exp
        return jwt.encode(data, key, algorithm='HS256')



WEBHOOK_CREATE_GQL = """
mutation webhookSubscriptionCreate($topic: WebhookSubscriptionTopic!, $webhookSubscription: WebhookSubscriptionInput!) {
  webhookSubscriptionCreate(topic: $topic, webhookSubscription: $webhookSubscription) {
    webhookSubscription {
      ...WebhookSubscription
    }
    userErrors {
      field
      message
    }
  }
}

mutation pubSubWebhookSubscriptionCreate($topic: WebhookSubscriptionTopic!, $webhookSubscription: PubSubWebhookSubscriptionInput!) {
  pubSubWebhookSubscriptionCreate(topic: $topic, webhookSubscription: $webhookSubscription) {
    webhookSubscription {
      ...WebhookSubscription
    }
    userErrors {
      field
      message
    }
  }
}

mutation eventBridgeWebhookSubscriptionCreate($topic: WebhookSubscriptionTopic!, $webhookSubscription: EventBridgeWebhookSubscriptionInput!) {
  eventBridgeWebhookSubscriptionCreate(topic: $topic, webhookSubscription: $webhookSubscription) {
    webhookSubscription {
      ...WebhookSubscription
    }
    userErrors {
      field
      message
    }
  }
}

fragment WebhookSubscription on WebhookSubscription {
  id
  topic
  format
  endpoint {
    __typename
    ... on WebhookHttpEndpoint {
      callbackUrl
    }
    ... on WebhookEventBridgeEndpoint {
      arn
    }
    ... on WebhookPubSubEndpoint {
      pubSubProject
      pubSubTopic
    }
  }
}
"""



from re import search


def domain_to_storename(domain: str) -> str:
    result = search(r'(https:\/\/)?([^.]+)\.myshopify\.com[\/]?', domain)
    if result:
        return result.group(2)

    raise ValueError(f'{domain} is not a shopify domain')


def store_domain(shop: str) -> str:
    """Conversion of a shop's subdomain or complete or incomplete url into a complete url."""
    result = search(r'^(https:\/\/)?([a-z1-9\-]+)(\.myshopify\.com[\/]?)?$', shop.lower())
    if not result:
        raise ValueError(f'{shop} is not a shopify shop')

    domain = result.group(3) or '.myshopify.com'
    storename = result.group(2)

    return f'{storename}{domain}'



from typing import Any

from httpx import AsyncClient


class HTTPClient(AsyncClient):
    __instance: Any = None

    def __new__(cls):
        if HTTPClient.__instance is None:
            HTTPClient.__instance = AsyncClient()
        return HTTPClient.__instance

    @classmethod
    async def close(cls):
        # graceful shutdown
        await HTTPClient.__instance.aclose()



from enum import Enum
from typing import List, Optional, Union

from pydantic import BaseModel

from spylib.admin_api import OfflineTokenABC
from spylib.constants import UTF8ENCODING
from spylib.exceptions import ShopifyGQLUserError
from spylib.hmac import validate as validate_hmac
from spylib.webhook.graphql_queries import WEBHOOK_CREATE_GQL


class WebhookTopic(Enum):
    ORDERS_CREATE = 'ORDERS_CREATE'


class WebhookResponse(BaseModel):
    id: str


class WebhookCreate(Enum):
    HTTP = 'webhookSubscriptionCreate'
    EVENT_BRIDGE = 'eventBridgeWebhookSubscriptionCreate'
    PUB_SUB = 'pubSubWebhookSubscriptionCreate'


def validate(data: Union[str, bytes], hmac_header: str, api_secret_key: str) -> bool:
    data_str: str
    if isinstance(data, bytes):
        data_str = data.decode(UTF8ENCODING)
    else:
        data_str = data
    try:
        validate_hmac(
            secret=api_secret_key, sent_hmac=hmac_header, message=data_str, use_base64=True
        )
    except ValueError:
        return False
    return True


async def create_http(
    offline_token: OfflineTokenABC,
    topic: Union[WebhookTopic, str],
    callback_url: str,
    include_fields: Optional[List[str]] = None,
    metafield_namespaces: Optional[List[str]] = None,
    private_metafield_namespaces: Optional[List[str]] = None,
) -> WebhookResponse:
    """Creates a HTTP webhook subscription.

    Uses graphql to subscribe to a webhook and associate it with an HTTP endpoint.
    """
    variables = {
        'topic': topic.value if isinstance(topic, WebhookTopic) else topic,
        'webhookSubscription': {
            'format': 'JSON',
            'includeFields': include_fields,
            'metafieldNamespaces': metafield_namespaces,
            'privateMetafieldNamespaces': private_metafield_namespaces,
            'callbackUrl': callback_url,
        },
    }
    res = await offline_token.execute_gql(
        query=WEBHOOK_CREATE_GQL,
        operation_name=WebhookCreate.HTTP.value,
        variables=variables,
    )
    webhook_create = res.get(WebhookCreate.HTTP.value, None)
    if webhook_create and webhook_create.get('userErrors'):
        raise ShopifyGQLUserError(res)
    return WebhookResponse(id=webhook_create['webhookSubscription']['id'])


async def create_event_bridge(
    offline_token: OfflineTokenABC,
    topic: Union[WebhookTopic, str],
    arn: str,
    include_fields: Optional[List[str]] = None,
    metafield_namespaces: Optional[List[str]] = None,
    private_metafield_namespaces: Optional[List[str]] = None,
) -> WebhookResponse:
    """Creates an Amazon EventBridge webhook subscription.

    Uses graphql to subscribe to a webhook and associated it with
    and AWS Event Bridge with ARN (Amazon Resource Name)
    """
    variables = {
        'topic': topic.value if isinstance(topic, WebhookTopic) else topic,
        'webhookSubscription': {
            'format': 'JSON',
            'includeFields': include_fields,
            'metafieldNamespaces': metafield_namespaces,
            'privateMetafieldNamespaces': private_metafield_namespaces,
            'arn': arn,
        },
    }
    res = await offline_token.execute_gql(
        query=WEBHOOK_CREATE_GQL,
        operation_name=WebhookCreate.EVENT_BRIDGE.value,
        variables=variables,
    )
    webhook_create = res.get(WebhookCreate.EVENT_BRIDGE.value, None)
    if webhook_create and webhook_create.get('userErrors'):
        raise ShopifyGQLUserError(res)
    return WebhookResponse(id=webhook_create['webhookSubscription']['id'])


async def create_pub_sub(
    offline_token: OfflineTokenABC,
    topic: Union[WebhookTopic, str],
    pub_sub_project: str,
    pub_sub_topic: str,
    include_fields: Optional[List[str]] = None,
    metafield_namespaces: Optional[List[str]] = None,
    private_metafield_namespaces: Optional[List[str]] = None,
) -> WebhookResponse:
    """Creates a Google Cloud Pub/Sub webhook subscription.

    Uses graphql to subscribe to a webhook and associate it with a Google PubSub endpoint
    """
    variables = {
        'topic': topic.value if isinstance(topic, WebhookTopic) else topic,
        'webhookSubscription': {
            'format': 'JSON',
            'includeFields': include_fields,
            'metafieldNamespaces': metafield_namespaces,
            'privateMetafieldNamespaces': private_metafield_namespaces,
            'pubSubProject': pub_sub_project,
            'pubSubTopic': pub_sub_topic,
        },
    }
    res = await offline_token.execute_gql(
        query=WEBHOOK_CREATE_GQL,
        operation_name=WebhookCreate.PUB_SUB.value,
        variables=variables,
    )
    webhook_create = res.get(WebhookCreate.PUB_SUB.value, None)
    if webhook_create and webhook_create.get('userErrors'):
        raise ShopifyGQLUserError(res)
    return WebhookResponse(id=webhook_create['webhookSubscription']['id'])



from inspect import isawaitable
from typing import Awaitable, Callable, Optional

from .exchange_token import exchange_offline_token, exchange_online_token
from .models import OfflineTokenModel, OnlineTokenModel
from .tokens import OAuthJWT
from .validations import validate_callback, validate_oauthjwt


async def process_callback(
    shop: str,
    timestamp: int,
    query_string: str,
    api_secret_key: str,
    api_key: str,
    state: str,
    private_key: str,
    code: str,
    post_install: Callable[[str, OfflineTokenModel], Optional[Awaitable]],
    post_login: Optional[Callable[[str, OnlineTokenModel], Optional[Awaitable]]],
) -> OAuthJWT:
    validate_callback(
        shop=shop,
        timestamp=timestamp,
        query_string=query_string,
        api_secret_key=api_secret_key,
    )
    oauthjwt: OAuthJWT = validate_oauthjwt(token=state, shop=shop, jwt_key=private_key)

    if not oauthjwt.is_login:
        await process_install_callback(
            oauthjwt=oauthjwt,
            shop=shop,
            code=code,
            api_key=api_key,
            api_secret_key=api_secret_key,
            post_install=post_install,
        )
        return oauthjwt

    await process_login_callback(
        oauthjwt=oauthjwt,
        shop=shop,
        code=code,
        api_key=api_key,
        api_secret_key=api_secret_key,
        post_login=post_login,
    )

    return oauthjwt


async def process_install_callback(
    oauthjwt: OAuthJWT,
    shop: str,
    code: str,
    api_key: str,
    api_secret_key: str,
    post_install: Callable[[str, OfflineTokenModel], Optional[Awaitable]],
):
    offline_token = await exchange_offline_token(
        shop=shop,
        code=code,
        api_key=api_key,
        api_secret_key=api_secret_key,
    )

    # Await if the provided function is async
    if isawaitable(pi_return := post_install(oauthjwt.storename, offline_token)):
        await pi_return  # type: ignore


async def process_login_callback(
    oauthjwt: OAuthJWT,
    shop: str,
    code: str,
    api_key: str,
    api_secret_key: str,
    post_login: Optional[Callable[[str, OnlineTokenModel], Optional[Awaitable]]],
):
    # Get the online token from Shopify
    online_token = await exchange_online_token(
        shop=shop,
        code=code,
        api_key=api_key,
        api_secret_key=api_secret_key,
    )

    # Await if the provided function is async
    if post_login:
        if isawaitable(pl_return := post_login(oauthjwt.storename, online_token)):
            await pl_return  # type: ignore
